#include <iostream>
#include "fixedFocusMethod.h"

#define sq(x) ((x) * (x))

/**
 * @file ΥˡΥإå
 * @brief ΥˡǾΥ׻
 * @param F ù
 * @param f ξΥ
 * @param f1 f2 Υ
 * @author Kento YAMADA
 * @date 2009.09.30
 */
void fixedFocusMethod(const CvMat *F,
                     const double f,
                     double &f1,
                     double &f2) {

  /* {{{ Υ׻ɬפʥѥ᡼η׻ */
  CvMat *k = cvCreateMat(3, 1, CV_64FC1);
  CvMat *vecTmp = cvCreateMat(3, 1, CV_64FC1);
  CvMat *matTmp = cvCreateMat(3, 3, CV_64FC1);
  cvmSet(k, 0, 0, 0.0);
  cvmSet(k, 1, 0, 0.0);
  cvmSet(k, 2, 0, 1.0);

	cvmMul(F, k, vecTmp);
  double Fk  = cvNorm(vecTmp);
  double Fk2 = sq(Fk);
 
	cvTranspose(F, matTmp);
  cvmMul(matTmp, k, vecTmp);
  double Ftk  = cvNorm(vecTmp);
  double Ftk2 = sq(Ftk);

	cvmMulTransposed(F, matTmp, 0);
  cvmMul(matTmp, F, matTmp);
  cvmMul(matTmp, k, vecTmp);
  double kFFtFk = cvDotProduct(k, vecTmp);

  cvmMul(F, k, vecTmp);
  double kFk = cvDotProduct(k, vecTmp);

  cvmMulTransposed(F, matTmp, 0);
  double FFt = cvNorm(matTmp);
  cvmMul(matTmp, k, vecTmp);
  double FFtk2 = sq(cvNorm(vecTmp));

  cvmMulTransposed(F, matTmp, 1);
  cvmMul(matTmp, k, vecTmp);
  double FtFk2 = sq(cvNorm(vecTmp));

  double a1 = 0.5 * sq(sq(kFk));
  double a2 = sq(kFk) * (Ftk2 + Fk2);
  double a3 = (0.5 * sq(Ftk2 - Fk2)) + kFk * ((4.0 * kFFtFk) - (kFk * sq(cvNorm(F))));
  double a4 = (2.0 * (FFtk2 + FtFk2)) - ((Ftk2 + Fk2) * sq(cvNorm(F)));
  double a5 = sq(FFt) - (0.5 * sq(sq(cvNorm(F))));
  /* }}} */

  /* {{{ Ƚ */
  double xi;
  double condition = 0.1 * ((Fk > Ftk) ? Ftk : Fk) / f;
  if(fabs(kFk) < condition) {
    xi = -a4 / (2.0 * a3);
  }
  /* }}} */

  /* {{{ 3 ¿༰η׻ */
  CvMat *coeffs = cvCreateMat(4, 1, CV_64FC1);
  CvMat *roots = cvCreateMat(3, 1, CV_64FC1);
  cvmSet(coeffs, 0, 0, 4.0 * a1);
  cvmSet(coeffs, 1, 0, 3.0 * a2);
  cvmSet(coeffs, 2, 0, 2.0 * a3);
  cvmSet(coeffs, 3, 0, a4);
  cvSolveCubic(coeffs, roots);
  double xiArray[3];
  int zeroNum = 0;
  int nonZero;
  for(int i = 0; i < 3; i++) {
    xiArray[i] = cvmGet(roots, i, 0);
  }

  for(int i = 0; i < 3; i++) {
    for(int j = i+1; j < 3; j++) {
      if(xiArray[i] < xiArray[j]) {
        std::swap(xiArray[i], xiArray[j]);
      }
    }
  }

  for(int i = 0; i < 3; i++) {
    if(xiArray[i] == 0.0) {
      zeroNum++;
    } else {
      nonZero = i;
    }
  }

  if(zeroNum == 0) {
    double Kxi3 = (a1 * sq(sq(xiArray[2]))) + (a2 * sq(xiArray[2]) * xiArray[2]) + (a3 * sq(xiArray[2])) + (a4 * xiArray[2]) + a5;
    double Kxi1 = (a1 * sq(sq(xiArray[0]))) + (a2 * sq(xiArray[0]) * xiArray[0]) + (a3 * sq(xiArray[0])) + (a4 * xiArray[0]) + a5;
    if((xiArray[2] < -1) || (Kxi3 < 0.0) || (Kxi1 < Kxi3)) {
      xi = xiArray[0];
	  } else if((0.0 < Kxi3) && (Kxi3 < Kxi1)) {
      xi = xiArray[2];
    } else {
      xi = -1;
    }
  } else if(zeroNum == 2) {
    xi = xiArray[nonZero];
  } else {
    xi = -1;
	}
  /* }}} */

  /* {{{ Ѥ CvMat ʤɤγ */
  cvReleaseMat(&k);
  cvReleaseMat(&vecTmp);
  cvReleaseMat(&matTmp);
  cvReleaseMat(&coeffs);
  cvReleaseMat(&roots);
  /* }}} */

  /* {{{ Υӽ */
  if(xi < -1) {
    std::cout << "# ERROR : Υ" << std::endl;
    f1 = f2 = f;
  } else {
    f1 = f2 = f / sqrt(1.0 + xi);
  }
  /* }}} */
}
